package
{
	import com.adobe.utils.*;

	import flash.display.*;
	import flash.display3D.*;
	import flash.display3D.textures.*;
	import flash.events.*;
	import flash.geom.*;
	import flash.text.*;
	import flash.utils.*;
	
	public class Picking extends Sprite
	{
		/** Number of degrees to rotate per millisecond */
		private static const ROTATION_SPEED:Number = 1;
		
		/** Axis to rotate about */
		private static const ROTATION_AXIS:Vector3D = new Vector3D(0, 1, 0);
		
		/** Distance between spheres */
		private static const SPHERE_SPACING:Number = 1.5;
		
		[Embed(source="earth.jpg")]
		private static const TEXTURE:Class;
		
		/** Temporary matrix to avoid allocation during drawing */
		private static const TEMP_DRAW_MATRIX:Matrix3D = new Matrix3D();
		
		/** 3D context to draw with */
		private var context3D:Context3D;
		
		/** Shader program to draw with */
		private var program:Program3D;
		
		/** Texture of all spheres */
		private var texture:Texture;
		
		/** Camera viewing the 3D scene */
		private var camera:Camera3D;
		
		/** Spheres to draw */
		private var spheres:Vector.<Sphere3D> = new Vector.<Sphere3D>();
		
		/** Current rotation of all spheres (degrees) */
		private var rotationDegrees:Number = 0;
		
		/** Framerate display */
		private var fps:TextField = new TextField();
		
		/** Last time the framerate display was updated */
		private var lastFPSUpdateTime:uint;
		
		/** Time when the last frame happened */
		private var lastFrameTime:uint;
		
		/** Number of frames since the framerate display was updated */
		private var frameCount:uint;
		
		/** 3D rendering driver display */
		private var driver:TextField = new TextField();
		
		/** Sphere that was picked with the last click or null if none was picked */
		private var picked:Sphere3D;
		
		/**
		* Entry point
		*/
		public function Picking()
		{
			stage.align = StageAlign.TOP_LEFT;
			stage.scaleMode = StageScaleMode.NO_SCALE;
			stage.frameRate = 60;
			
			var stage3D:Stage3D = stage.stage3Ds[0];
			stage3D.addEventListener(Event.CONTEXT3D_CREATE, onContextCreated);
			stage3D.requestContext3D(Context3DRenderMode.AUTO);
		}
		
		protected function onContextCreated(ev:Event): void
		{
			// Setup context
			var stage3D:Stage3D = stage.stage3Ds[0];
			stage3D.removeEventListener(Event.CONTEXT3D_CREATE, onContextCreated);
			context3D = stage3D.context3D;            
			context3D.configureBackBuffer(
				stage.stageWidth,
				stage.stageHeight,
				0,
				true
			);
			
			// Setup camera
			camera = new Camera3D(
				0.1, // near
				100, // far
				stage.stageWidth / stage.stageHeight, // aspect ratio
				40*(Math.PI/180), // vFOV
				0, 2, 5, // position
				0, 2, 0, // target
				0, 1, 0 // up dir
			);
			
			// Setup UI
			fps.background = true;
			fps.backgroundColor = 0xffffffff;
			fps.autoSize = TextFieldAutoSize.LEFT;
			fps.text = "Getting FPS...";
			addChild(fps);
			
			driver.background = true;
			driver.backgroundColor = 0xffffffff;
			driver.text = "Driver: " + context3D.driverInfo;
			driver.autoSize = TextFieldAutoSize.LEFT;
			driver.y = fps.height;
			addChild(driver);
			
			var assembler:AGALMiniAssembler = new AGALMiniAssembler();
			
			// Vertex shader
			var vertSource:String = "m44 op, va0, vc0\nmov v0, va1\n";
			assembler.assemble(Context3DProgramType.VERTEX, vertSource);
			var vertexShaderAGAL:ByteArray = assembler.agalcode;
			
			// Fragment shader
			var fragSource:String = "tex oc, v0, fs0 <2d,linear,mipnone>";
			assembler.assemble(Context3DProgramType.FRAGMENT, fragSource);
			var fragmentShaderAGAL:ByteArray = assembler.agalcode;
			
			// Shader program
			program = context3D.createProgram();
			program.upload(vertexShaderAGAL, fragmentShaderAGAL);
			
			// Setup textures
			var bmd:BitmapData = (new TEXTURE() as Bitmap).bitmapData;
			texture = context3D.createTexture(
				bmd.width,
				bmd.height,
				Context3DTextureFormat.BGRA,
				true
			);
			texture.uploadFromBitmapData(bmd);
			
			for (var row:int = 0; row < 3; ++row)
			{
				for (var col:int = 0; col < 3; ++col)
				{
					spheres.push(
						new Sphere3D(
							context3D,
							10,
							10,
							col*SPHERE_SPACING,
							row*SPHERE_SPACING,
							0
						)
					);
				}
			}
			
			// Start the simulation
			stage.addEventListener(MouseEvent.CLICK, onMouseClick);
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}
		
		private function pick(x:Number, y:Number): Sphere3D
		{
			// Normalize the click
			var xUnit:Number = (x * 2 / stage.stageWidth) - 1.0;
			var yUnit:Number = -((y * 2 / stage.stageHeight) - 1.0);
			
			// Make the picking ray
			var rayOrigin:Vector3D = new Vector3D();
			var rayDir:Vector3D = new Vector3D();
			camera.getPickingRay(xUnit, yUnit, rayOrigin, rayDir);
			
			// Test the picking ray against all renderables
			var intersectionDistance:Number = Infinity;
			var picked:Sphere3D;
			for each (var sphere:Sphere3D in spheres)
			{
				// Compute intersection distance and reject non-intersections
				// and bounding volumes we're inside
				var distance:Number = sphere.intersectRay(rayOrigin, rayDir);
				if (!(distance > 0) || distance > intersectionDistance)
				{
					continue;
				}
				
				intersectionDistance = distance;
				picked = sphere;
			}
			
			return picked;
		}
		
		private function onMouseClick(ev:MouseEvent): void
		{
			// Only accept clicks on the 3D simulation
			if (ev.target != stage)
			{
				return;
			}
			
			picked = pick(ev.stageX, ev.stageY);
		}
		
		private function onEnterFrame(ev:Event): void
		{
			// Set up rendering
			context3D.setProgram(program);
			context3D.setTextureAt(0, texture);
			context3D.clear(0.5, 0.5, 0.5);
			
			// Draw spheres
			var worldToClip:Matrix3D = camera.worldToClipMatrix;
			var drawMatrix:Matrix3D = TEMP_DRAW_MATRIX;
			for each (var sphere:Sphere3D in spheres)
			{
				context3D.setVertexBufferAt(0, sphere.positions, 0, Context3DVertexBufferFormat.FLOAT_3);
				context3D.setVertexBufferAt(1, sphere.texCoords, 0, Context3DVertexBufferFormat.FLOAT_2);
				
				sphere.modelToWorld.copyToMatrix3D(drawMatrix);
				if (sphere != picked)
				{
					drawMatrix.appendRotation(rotationDegrees, ROTATION_AXIS);
				}
				drawMatrix.prepend(worldToClip);
				context3D.setProgramConstantsFromMatrix(
					Context3DProgramType.VERTEX,
					0,
					drawMatrix,
					false
				);
				context3D.drawTriangles(sphere.tris);
			}
			context3D.present();
			
			rotationDegrees += ROTATION_SPEED;
			
			// Update stat displays
			frameCount++;
			var now:int = getTimer();
			var elapsed:int = now - lastFPSUpdateTime;
			if (elapsed > 1000)
			{
				var framerateValue:Number = 1000 / (elapsed / frameCount);
				fps.text = "FPS: " + framerateValue.toFixed(1);
				lastFPSUpdateTime = now;
				frameCount = 0;
			}
			lastFrameTime = now;
		}
	}
}
